##############################################################################
#   Environment Sensor V2.1
#   - Temperature, Pressure, Humidity
#       - Gas Oxidising, Reducing, NH3
#       - Lux, Proximity, Noise
#   - Partical Detection 1.0 2.5 10 microns
#
#   Uses Enviro+ Pi Hat from Pimoroni and PMS5003 Particle Sensor
#
#   NOTE: Due to BME280 being in the Hat in close proximity to the Pi,
#         Temperature and Humidity readings are pretty much useless.
#         They only indicate the conditions in the enclosure.
#
#
# enviro2.py
#
# Display sensor readings on the LCD on the Enviro+ board.
# Use the proximity sensor to detect when to display and change display.
#
#
# James Grigg - 16 Nov 2020
#
#    16 Nov 2020   Modification of GitHub EnviroPlusWeb application, display data as a graph
#                  Seems better than what I did in V2.0
#                  Sensors are read every 2 to 3 seconds averaged after 5 miutes and saved to file
#    17 Nov 2020   Run Sensors LCD & Terminal as threads.
#
#
#
##############################################################################
#!/usr/bin/env python3
import os
import datetime
import math
import sounddevice
import numpy
import ST7735
import board
import busio
import adafruit_bmp280
import base64
import multiprocessing
import board
import busio
import adafruit_bmp280
import logging
import threading
import json
import RPi.GPIO as IO

from time import sleep, time, asctime, localtime, strftime, gmtime
from pms5003 import PMS5003, ReadTimeoutError as pmsReadTimeoutError
from math import ceil, floor
from bme280 import BME280
from enviroplus import gas
from PIL import Image, ImageDraw, ImageFont
from fonts.ttf import RobotoMedium as UserFont
from subprocess import call
from importlib import import_module
from flask import Flask, render_template, url_for, request, Response

try:
    from ltr559 import LTR559
    ltr559 = LTR559()
except ImportError:
    import ltr559
try:
  from smbus2 import SMBus
except ImportError:
  from smbus import SMBus

#########################################################################
# Interface setup, BME280 address 0x76 (118) BMP280 address 0x77 (119)
#########################################################################
i2c = busio.I2C(board.SCL, board.SDA)
bmp280 = adafruit_bmp280.Adafruit_BMP280_I2C(i2c,119)

bus = SMBus(1)
bme280 = BME280(i2c_dev=bus)

pms5003 = PMS5003(device='/dev/ttyAMA0',baudrate=9600,pin_enable=22,pin_reset=27)

IO.setmode(IO.BCM)   # Set pin numbering
IO.setup(4,IO.OUT)   # Fan controller on GPIO 4
pwm = IO.PWM(4,1000) # PWM frequency
pwm.start(100)       # Duty cycle
##############################################################################
# LCD Setup
##############################################################################
st7735 = ST7735.ST7735(port=0,cs=1,dc=9,backlight=12,rotation=270,spi_speed_hz=10000000)
st7735.begin()

WIDTH = st7735.width
HEIGHT = st7735.height

img = Image.new('RGB', (WIDTH, HEIGHT), color=(0, 0, 0))
draw = ImageDraw.Draw(img)
x_offset = 2
y_offset = 2
#units = ["°C","%","mBar","Lux","kΩ","kΩ","kΩ","/100cl","/100cl","/100cl"]

font_size = 20      # 15 characters
#smallfont = ImageFont.truetype(UserFont, 10)
#font = ImageFont.truetype(UserFont, font_size)
#font = ImageFont.truetype('/home/pi/Perfect DOS VGA 437 Win.ttf', font_size)
font = ImageFont.truetype('/home/pi/Roboto-Regular.ttf', font_size)
text_colour = (255, 255, 255)
back_colour = (0, 170, 170)
# Display is 15 character x 3 lines. Readable using this font.

#########################################################################
# Variables
#########################################################################
bmp280.sea_level_pressure = 1013.25

record = {}                                  # sensor readings now
data = []                                    # current sensor readings for up to last 5 minutes
days = []                                    # holds all the saved data
samples = 300                                # Each file record is 5 minutes, 60 x 5 = 300 seconds
samples_per_day = 24 * 3600 // samples       # // rounds down and truncates

Time_Start = datetime.datetime.now()
Seconds_Start = Time_Start.timestamp()

#########################################################################
# Set up web application
#########################################################################
app = Flask(__name__)
app.config['TEMPLATES_AUTO_RELOAD'] = True

log = logging.getLogger("werkzeug")
log.disabled = True
run_flag = True

#########################################################################
# Create a filename from year and day number, ie 365 days in year
#########################################################################
def filename(t):
    return strftime("data/%Y_%j", localtime(t))

#########################################################################
# Centre Message on line
#########################################################################
def Centre_MSG(m):
    size_x, size_y = draw.textsize(m, font)
    x = (WIDTH - size_x) / 2
    return x
#########################################################################
# Display LCD Message - 3 lines of 15 characters, centered or not
#########################################################################
def Display_MSG(L1,L2,L3,c):

    x = 2    # leave 2 pixels for non centered message
    draw.rectangle((0, 0, 160, 80), back_colour)
    if c == True:
        x = Centre_MSG(L1)
    draw.text((x,  1), L1, font=font, fill=text_colour)
    if c == True:
        x = Centre_MSG(L2)
    draw.text((x, 26), L2, font=font, fill=text_colour)
    if c == True:
        x = Centre_MSG(L3)
    draw.text((x, 52), L3, font=font, fill=text_colour)
    st7735.display(img)

    return
#########################################################################
# Display text and data
#########################################################################
def Display_Data(L1T,L2T,L3T,L1D,L2D,L3D):

    draw.rectangle((0, 0, 160, 80), back_colour)
    draw.text((2,  1),    L1T, font=font, fill=text_colour)
    x,y = draw.textsize(L1D, font)
    draw.text((156-x,  1), L1D, font=font, fill=text_colour)
    draw.text((2, 26),    L2T, font=font, fill=text_colour)
    x,y = draw.textsize(L2D, font)
    draw.text((156-x, 26), L2D, font=font, fill=text_colour)
    draw.text((2, 52),    L3T, font=font, fill=text_colour)
    x,y = draw.textsize(L3D, font)
    draw.text((156-x, 52), L3D, font=font, fill=text_colour)
    st7735.display(img)

    return
#########################################################################
# LCD Display Screens
#########################################################################
def Info():
    Time_Now = datetime.datetime.now()
    Display_MSG("Environment","Monitor",Time_Now.strftime("%d/%m/%y  %H:%M"),True)
    return

def LCD_Temp_Humid_Pressure():
    Display_Data(" Temp."," Humid."," Press.",
                 "{:8.02f}".format(record['temp']),
                 "{:8.02f}".format(record['humi']),
                 "{:8.02f}".format(record['pres']))
    return

def LCD_Temp_Press_BMP280():
    Display_Data(" Temp."," Press.","BMP280",
                 "{:8.02f}".format(record['tempP']),
                 "{:8.02f}".format(record['presP'])," ")
    return

def LCD_PM():
    Display_Data(" PM 10.0"," PM  2.5"," PM  1.0",
                 "{:5.0f}".format(record['pm100']),
                 "{:5.0f}".format(record['pm25']),
                 "{:5.0f}".format(record['pm10']))
    return

def LCD_GAS():
    Display_Data(" OX"," RED"," NH3",
                 "{:8.0f}".format(record['oxi']),
                 "{:8.0f}".format(record['red']),
                 "{:8.0f}".format(record['nh3']))
    return

def LCD_Noise_Light():
    Display_Data(" Lux"," Noise"," ",
                 "{:8.02f}".format(record['lux']),
                 "{:8.02f}".format(record['noise'])," ")
    return
#########################################################################
# LCD Screen selector
#########################################################################
def LCD_Screen_Selector(s):
    switcher = {
        0: Info,
        1: LCD_Temp_Humid_Pressure,
        2: LCD_Temp_Press_BMP280,
        3: LCD_PM,
        4: LCD_GAS,
        5: LCD_Noise_Light
    }
    st7735.set_backlight(12)
    func = switcher.get(s, lambda: "Invalid Screen")
    func()
    return
#########################################################################
# LCD Display
#########################################################################
def LCD():
    prox_trig = False
    LCD_OFF = False
    LCD_Screen = 0

    text_colour = (255, 255, 255)
    back_colour = (0, 170, 170)
    Display_MSG("Environment","Monitor",Time_Start.strftime("%d/%m/%y  %H:%M"),True)
    LCD_Display_Timer = datetime.datetime.now().timestamp() + 30

    while run_flag:
        if ltr559.get_proximity() > 600:
            if prox_trig == False:
                if LCD_OFF == True:
                    LCD_Screen = 0
                LCD_Screen_Selector(LCD_Screen)
                LCD_Display_Timer = datetime.datetime.now().timestamp() + 30
                LCD_OFF = False
                LCD_Screen += 1
                if LCD_Screen > 5:
                    LCD_Screen = 0
                prox_trig = True
            else:
                prox_trig = False

        if LCD_Display_Timer < datetime.datetime.now().timestamp():   # LCD turns off after 30 seconds
            st7735.set_backlight(0)                                   # if no new screen
            draw.rectangle((0, 0, 160, 80), back_colour)
            st7735.display(img)
            LCD_OFF = True

        sleep(0.3)

#########################################################################
#########################################################################
# Converts Time String into integer
#########################################################################
def record_time(r):
    t = r['time'].split()[3].split(':')
    return int(t[0]) * 60 + int(t[1])

#########################################################################
#  Adds Sensor reads to list for 5 minute period
#########################################################################
def add_record(day, record):
    if record_time(record) > 0:                   # If not the first record of the day
        while len(day) == 0 or record_time(day[-1]) < record_time(record) - samples // 60: # Is there a gap
            if len(day):
                filler = dict(day[-1])            # Duplicate the last record to forward fill
                t = record_time(filler) + samples // 60
            else:
                filler = dict(record)             # Need to back fill
                t = 0                             # Only happens if the day is empty so most be the first entry
            old_time = filler["time"]             # Need to fix the time field
            colon_pos = old_time.find(':')
            filler["time"] = old_time[:colon_pos - 2] + ("%02d:%02d" % (t / 60, t % 60)) + old_time[colon_pos + 3:]
            day.append(filler)
    day.append(record)

#########################################################################
# Averge Sensor readings for save to file, 5 minute period
#########################################################################
def sum_data(data):
    totals = {"time" : data[0]["time"]}
    keys = list(data[0].keys())
    keys.remove("time")
    for key in keys:
        totals[key] = 0
    for d in data:
        for key in keys:
            totals[key] += d[key]
    count = float(len(data))
    for key in keys:
        totals[key] = round(totals[key] / count, 1)
    return totals

#########################################################################
# Read Sensors
#########################################################################
def read_data(time):

    temperature = bme280.get_temperature()
    pressure = bme280.get_pressure()
    humidity = bme280.get_humidity()

    tempBMP280 = bmp280.temperature-2
    pressBMP280 =  bmp280.pressure

    lux = ltr559.get_lux()
    gases = gas.read_all()

    Rec = sounddevice.rec(8000,16000,blocking=True,channels=1,dtype='float64')
    RecAbs = numpy.abs(Rec)
    RecMax = numpy.max(RecAbs)
    noise = 20 * math.log(RecMax/1)                      # convert level to dB

    while True:
        try:
            particles = pms5003.read()
            break
        except RuntimeError as e:
            print("Particle read failed:", e.__class__.__name__)
            if not run_flag:
                raise e
            pms5003.reset()
            sleep(30)

    pm100 = particles.pm_per_1l_air(10.0)
    pm50 = particles.pm_per_1l_air(5.0)
    pm25 = particles.pm_per_1l_air(2.5)
    pm10 = particles.pm_per_1l_air(1.0)
    pm5 = particles.pm_per_1l_air(0.5)
    pm3 = particles.pm_per_1l_air(0.3)

    record = {
        'time' : asctime(localtime(time)),
        'temp' : round(temperature,1),
        'humi' : round(humidity, 1),
        'pres' : round(pressure,1),
        'tempP': round(tempBMP280,1),
        'presP': round(pressBMP280,1),
        'lux'  : round(lux),
        'noise': round(noise,1),
        'oxi'  : round(gases.oxidising / 1000, 1),
        'red'  : round(gases.reducing / 1000),
        'nh3'  : round(gases.nh3 / 1000),
        'pm03' : pm3,
        'pm05' : pm5,
        'pm10' : pm10,
        'pm25' : pm25,
        'pm50' : pm50,
        'pm100': pm100,
    }
    return record
#########################################################################
# Reads Sensors and places in list, runs as separate thead
#########################################################################
def Sensors():
    global record, data

    sleep(2)
    last_file = None
    while run_flag:
        t = int(floor(time()))
        record = read_data(t)
        data = data[-(samples - 1):] + [record]                  # Keep five minutes
        if t % samples == samples - 1 and len(data) == samples:  # At the end of a 5 minute period?
            totals = sum_data(data)
            fname = filename(t - (samples - 1))
            with open(fname, "a+") as f:
                f.write(json.dumps(totals) + '\n')
            if not days or (last_file and last_file != fname):
                days.append([])
            last_file = fname
            add_record(days[-1], totals)                         # Add to today, filling any gap from last reading
                                                                 # if it had been stopped
        sleep(max(t + 1 - time(), 0.1))

#########################################################################
#########################################################################
# Startup Message on terminal
#########################################################################
def Start_MSG():
    print (" ")
    print ("       Environment Monitor - V2")
    print (" --  Cover Light Sensor to cycle LCD Screens--")
    print (" ")
    print ("Gas Sensor values are the resistance (k ohms) measurement of the sensor.")
    print ("Calibration is related to many factors and not possible for this project.")
    print ("Significant changes in these values would indicate Gas presence.")
    print ("The Gas Sensors can take up to an hour to warm up.")
    print ("BME280 Temperature and Humidity affected by closeness to Pi")
    print ("BMP280 Temperature is 2 to 4 degrees more than ambient")
    print (" ")
    print ("Start Time ",Time_Start.strftime("%d %b %Y  %H:%M:%S"))
    print (" ")

    return
#########################################################################
# Print Sensor Readings
#   Displaying 0.3 & 0.5 to indicate PMS5003 is working.
#########################################################################
def print_Sensors():
    global Time_Start

#    print (datetime.datetime.now().timestamp(),Time_Start)

    dt = datetime.datetime.now()
    t = int(dt.timestamp() - Seconds_Start)         # get seconds since start
    h = int(t / 3600)                            # calc hours minutes & seconds since start
    m = int((t - (h * 3600)) / 60)
    s = int(t - (h * 3600 + m *60))

    print ("|--------------------------------------|")
    print ("|       ",dt.strftime("%d %b %Y  %H:%M:%S"),"        |")
    print ("|----------------------------------------------------------------|")
    print ("| Temperature|  Pressure  |  Humidity  | Light Lux  | Noise dB   |")
    print ("|  {:8.02f}  |  {:8.02f}  | {:8.02f}   |  {:8.02f}  |  {:8.02f}  |"
         .format(record['temp'],record['pres'],record['humi'],record['lux'],record['noise']))
    print ("|------------|------------|--------------------------------------|")
    print ("| Temp BMP280| PressBMP280| Oxidising  |  Reducing  |     NH3    |")
    print ("|  {:8.02f}  | {:8.02f}   |  {:8.02f}  |  {:8.02f}  |  {:8.02f}  |"
         .format(record['tempP'],record['presP'],record['oxi'],record['red'],record['nh3']))
    print ("|------------|------------|--------------------------------------|")
    print ("|   PM 0.3   |   PM 0.5   |   PM 1.0   |   PM 2.5   |  PM 10.0   |")
    print ("|  {:7.0f}   |  {:7.0f}   |  {:7.0f}   |  {:7.0f}   |  {:7.0f}   |"
         .format(record['pm03'],record['pm05'],record['pm10'],record['pm25'],record['pm100']))
    print ("|----------------------------------------------------------------|")
    print ("   Run time = {:02d}:{:02d}:{:02d}".format(h,m,s))
    print (" ")

    return                               # return time for next display

#########################################################################
# Terminal Display
#########################################################################
def Terminal():

    Start_MSG()
    Print_Sensor_Timer = Seconds_Start + 60                               # display every minute
    while run_flag:
        if Print_Sensor_Timer < datetime.datetime.now().timestamp():
            print_Sensors()
            Print_Sensor_Timer = datetime.datetime.now().timestamp() + 60 # set time for next display
        sleep(1.0)

#########################################################################
#########################################################################
# Reduce large sample (year) to what can be displayed on web page
#########################################################################
def compress_data(ndays, nsamples):
    cdata = []
    for day in days[-(ndays + 1):]:
        for i in range(0, len(day), nsamples):
            cdata.append(sum_data(day[i : i + nsamples]))
    length = ndays * samples_per_day // nsamples
    return json.dumps(cdata[-length:])

# 300 @ 1s = 5m
# 288 @ 5m = 24h
# 336 @ 30m = 1w
# 372 @ 2h = 31d
# 365 @ 1d = 1y
#########################################################################
#  Web page access
#########################################################################
@app.route('/')
def index():
    return render_template('index.html')

@app.route('/readings')
def readings():
    arg = request.args["fan"]
    pwm.ChangeDutyCycle(int(arg))
    return render_template('readings.html', **record)

@app.route('/graph')
def graph():
    arg = request.args["time"]
    if arg == 'day':
        last2 = []
        for day in days[-2:]:
            last2 += day
        return json.dumps(last2[-samples_per_day:])
    if arg == 'week':
        return compress_data(7, 30 * 60 // samples)
    if arg == 'month':
        return compress_data(31, 120 * 60 // samples)
    if arg == 'year':
        return compress_data(365, samples_per_day)
    return json.dumps(data)

#########################################################################
#  Read all data that has been saved into a list for display on the graph
#########################################################################
def Get_Saved_Data():
    global days

    if not os.path.isdir('data'):                     # create folder if not present
        os.makedirs('data')
    files =  sorted(os.listdir('data'))               # get all the file names
    for f in files:
        day = []
        with open('data/' + f, 'r') as file:          # open a file
            for line in file.readlines():             # read each line
                record = json.loads(line)
                add_record(day, record)
        days.append(day)
    return

#########################################################################
#########################################################################
# Main Run Loop  Read the day data into a list from all the saved files
#                Start sensor reads in background
#                Start Flask web server as separate process
#########################################################################
if __name__ == '__main__':

    Sensors_thread = threading.Thread(target = Sensors)
    LCD_thread = threading.Thread(target = LCD)
    Term_thread = threading.Thread(target = Terminal)

    record = read_data(time())                        # first reading is not correct data
    Get_Saved_Data()                                  # get all saved data from all files
    Sensors_thread.start()                         # start sensor reads and saves
    LCD_thread.start()                                # start LCD display
    Term_thread.start()                                # start LCD display
    app.run(host='0.0.0.0', port=5000, threaded=True, debug=False)
# execution waits here till web process stopped

    st7735.set_backlight(0)                           # blank LCD before exit
    print (" ")
    print ("Waiting for background to quit")
    run_flag = False                                  # stops background thread from running
    Sensors_thread.join()                          # waits for background threads to stop
    LCD_thread.join()
    Term_thread.join()

#########################################################################
#########################################################################
#########################################################################
#########################################################################
#
#########################################################################
# Displays all the text on the 0.96" LCD
#          Note: if no saved files screen will be blank for first 5 mins
#          Extrememly tiny, very hard to read
#########################################################################
#def display_everything():
#    draw.rectangle((0, 0, WIDTH, HEIGHT), (0, 0, 0))
#    column_count = 2
#    variables = list(record.keys())
#    row_count = len(units) // column_count
#    last_values = days[-1][-1]
#    for i in range(len(units)):
#        variable = variables[i + 1]
#        data_value = record[variable]
#        last_value = last_values[variable]
#        unit = units[i]
#        x = x_offset + ((WIDTH // column_count) * (i // row_count))
#        y = y_offset + ((HEIGHT // row_count) * (i % row_count))
#        message = "{}: {:s} {}".format(variable[:4], str(data_value), unit)
#        tol = 1.01
#        rgb = (255, 0, 255) if data_value > last_value * tol  else (0, 255, 255) if data_value < last_value / tol else (0, 255, 0)
#        draw.text((x, y), message, font = smallfont, fill = rgb)
#    st7735.display(img)
#
#    try:
#        app.run(debug = False, host = '0.0.0.0', port = 5000, use_reloader = False)
#    except Exception as e:
#        print(e)
#        pass
#########################################################################
#########################################################################
#########################################################################
#########################################################################
#########################################################################
#########################################################################
#########################################################################